#include "window.h"
#include "ui_window.h"

#include <QApplication>
#include <QFileDialog>

#include <QFile>
#include <QXmlStreamReader>
#include <QXmlStreamAttribute>

#include <QDrag>
#include <QMimeData>
#include <QIcon>
#include <QGraphicsSceneDragDropEvent>
#include <QBitmap>
#include <QMetaProperty>

#include <cassert>

MainWindow * global_win = NULL;

/* ---------------------------------------------------------------------- */

/* parameter list */

int ParamList::number (QString key, int default_value)
{
    QString s = this->value (key);
    bool ok;
    int n = s.toInt (&ok);
    if (!ok)
        n = default_value;
    return n;
}

/* ---------------------------------------------------------------------- */

/* Shape */

const QString shape_format = "application/x-shape";

void shapeDragEnter (QGraphicsSceneDragDropEvent *event)
{
    const QMimeData * data = event->mimeData();
    if (data->hasColor() || data->hasFormat (shape_format))
        event->setAccepted (true);
}

void shapeDrop (QAbstractGraphicsShapeItem * shape, QGraphicsSceneDragDropEvent *event)
{
   const QMimeData * data = event->mimeData();
   if (data->hasColor())
   {
       QColor c =  data->colorData().value <QColor> ();
       if (event->proposedAction() == Qt::MoveAction) /* shift */
          shape->setPen (c);
       else
          shape->setBrush (c);
   }
   if (data->hasFormat (shape_format))
   {
       QString code = data->data (shape_format);
       if (global_win != NULL)
          global_win->addGraphicsItem (shape, code, event->proposedAction() == Qt::MoveAction);
   }

}

/* ---------------------------------------------------------------------- */

/* Rectangle */

void Rectangle::dragEnterEvent (QGraphicsSceneDragDropEvent *event)
{
    event->setAccepted (true);
    //shapeDragEnter (event);
}

void Rectangle::dropEvent (QGraphicsSceneDragDropEvent *event)
{
    shapeDrop (this, event);
}

Rectangle::Rectangle()
{
    setAcceptDrops (true);
}

/* ---------------------------------------------------------------------- */

/* Ellipse */

void Ellipse::dragEnterEvent (QGraphicsSceneDragDropEvent *event)
{
    shapeDragEnter (event);
}

void Ellipse::dropEvent (QGraphicsSceneDragDropEvent *event)
{
    shapeDrop (this, event);
}

Ellipse::Ellipse ()
{
    setAcceptDrops (true);
}

/* ---------------------------------------------------------------------- */

/* Color Button */

ColorButton::ColorButton (QString s)
  : name (s), color (QColor (s))
{
    setText (s);
    setToolTip (s);
    setIcon (createPixmap ());
}

void ColorButton::mousePressEvent(QMouseEvent *event)
{
    QDrag * drag = new QDrag (this);
    QMimeData * data = new QMimeData ();
    data->setText (name);
    data->setColorData (color);
    drag->setMimeData(data);
    drag->setPixmap (createPixmap (true));
    drag->start(Qt::CopyAction | Qt::MoveAction);
}

QPixmap ColorButton::createPixmap (bool white)
{
    QPixmap pixmap (24, 24);
    if (white)
       pixmap.fill (QColor (255, 255, 255, 255));
    else
       pixmap.fill (QColor (0, 0, 0, 0));

    QPainter painte (& pixmap);
    painte.setPen (color);
    painte.setBrush (color);
    painte.drawEllipse (2, 2, 20, 20);

    return pixmap;
};

/* ---------------------------------------------------------------------- */

/* Shape Button */

ShapeButton::ShapeButton (QString p_name, QString p_code)
  : name (p_name),
    code (p_code)
{
    code.replace ('\'', '\"');
    code = "<data>" + code + "</data>"; // add top level element
    setText (name);
    setToolTip (code);
}

void ShapeButton::mousePressEvent (QMouseEvent *event)
{
    QDrag * drag = new QDrag (this);
    QMimeData * data = new QMimeData ();
    data->setData (shape_format, code.toLatin1());
    drag->setMimeData (data);
    drag->start ();
}

/* ---------------------------------------------------------------------- */

/* create / update items */

QGraphicsItem * createItem (QString type, ParamList params)
{
    QAbstractGraphicsShapeItem * shape = NULL;
    if (type == "rectangle")
    {
        Rectangle * r = new Rectangle;
        shape = r;
        shape->setData (Qt::DisplayRole, "rectangle");

        int w = params.number ("width", 20);
        int h = params.number ("height", 20);
        r->setRect (0, 0, w, h);
    }
    else if (type == "ellipse")
    {
        Ellipse * e = new Ellipse;
        shape = e;
        shape->setData (Qt::DisplayRole, "ellipse");

        int w = params.number ("width", 20);
        int h = params.number ("height", 20);
        e->setRect (0, 0, w, h);
    }

    if (shape != NULL)
    {
        int x = params.number ("x", 0);
        int y = params.number ("y", 0);
        shape->setPos (x, y);

        QString s = params ["pen"];
        if (s != "")
           shape->setPen (QColor (s));

        s = params ["brush"];
        if (s != "")
           shape->setBrush (QColor (s));
     }

    return shape;
}

/* ---------------------------------------------------------------------- */

/* input / output */

void MainWindow::readXml (QXmlStreamReader & reader, QGraphicsItem * target)
{
    int level = 0;
    while (! reader.atEnd())
    {
        if (reader.isStartElement ())
        {
            if (level > 0) // ignore top level element
            {
                QString type = reader.name().toString();

                ParamList params;
                // for (QXmlStreamAttribute a : reader.attributes())
                for (auto a : reader.attributes())
                {
                   QString attr = a.name().toString();
                   QString value = a.value().toString();
                   params [attr] = value;
                }

                QGraphicsItem * item = createItem (type, params);
                assert (item != NULL);

                addIntoScene (target, item);

                target = item; // new branch
             }
             level ++;
        }
        else if (reader.isEndElement ())
        {
            level --;
            if (level > 0)
               target = target->parentItem ();
        }
        reader.readNext();
     }
}

void MainWindow::addIntoScene (QGraphicsItem * target, QGraphicsItem * item)
{
    assert (item != NULL);

    item->setFlag (QGraphicsItem::ItemIsMovable);
    item->setFlag (QGraphicsItem::ItemIsSelectable);

     if (target == NULL)
         // scene->addItem (item);
         item->setParentItem (top_item);
     else
         item->setParentItem (target);
}

void MainWindow::addGraphicsItem (QGraphicsItem * target, QString code, bool plus)
{
   if (plus)
   {
      if (target != NULL)
         target = target->parentItem ();
   }

   QXmlStreamReader reader (code);
   readXml (reader, target);
}

void MainWindow::readFile (QString fileName)
{
    QFile file (fileName);
    if (file.open (QFile::ReadOnly | QFile::Text ))
    {
        QXmlStreamReader reader (&file);
        readXml (reader, NULL);
        // ui->treeView->expandAll ();
    }
}

void MainWindow::writeFile (QString fileName)
{
    QFile file (fileName);
    if (file.open (QFile::WriteOnly | QFile::Text))
    {
        QXmlStreamWriter writer (&file);
        writer.setAutoFormatting (true);

        writer.writeStartDocument ();
        writer.writeStartElement ("data"); // top level element

        /*
        QList <QGraphicsItem *> items = scene->items();
        int cnt = items.length();
        for (int inx = 0; inx < cnt; inx ++)
        {
           QGraphicsItem * item = items [inx];
           if (item->parentItem () == NULL)
               writeItem (writer, item);
        }
        */

        QList <QGraphicsItem *> items = top_item->childItems ();
        for (QGraphicsItem * t : items)
        {
             writeItem (writer, t);
        }

        writer.writeEndElement ();
        writer.writeEndDocument ();
    }
}

void MainWindow::writeItem (QXmlStreamWriter & writer, QGraphicsItem * item)
{

   QObject * obj = dynamic_cast <QObject *> (item);
   if (obj != NULL)
   {
      const QMetaObject * cls = obj->metaObject ();
      assert (cls != NULL);

      QString type = "";
      if (Rectangle * r = dynamic_cast < Rectangle * > (item))
         type =  "rectangle";
      if (Ellipse * e = dynamic_cast < Ellipse * > (item))
         type = "ellipse";
      if (type == "")
         type = cls->className();
      writer.writeStartElement (type);

      int cnt = cls->propertyCount ();
      ui->tableWidget->setRowCount (cnt);
      for (int inx = 0; inx < cnt; inx ++)
      {
          QMetaProperty prop = cls->property (inx);
          QString attr = prop.name ();
          QString value = prop.read (obj).toString ();

          writer.writeAttribute (attr, value);
      }

      QList <QGraphicsItem *> items = item->childItems ();
      for (QGraphicsItem * t : items)
      {
           writeItem (writer, t);
      }

      writer.writeEndElement ();
   }
}

void MainWindow::on_actionOpen_triggered()
{
    QString s = QFileDialog::getOpenFileName (this, "Open XML file");
    if (s != "")
       readFile (s);
}

void MainWindow::on_actionSave_triggered()
{
    QString s = QFileDialog::getSaveFileName (this, "Save XML file");
    if (s != "")
       writeFile (s);
}

/* ---------------------------------------------------------------------- */

/* tree and property table */

void MainWindow::on_treeView_clicked(const QModelIndex &index)
{
    QGraphicsItem * item = static_cast <QGraphicsItem * > (index.internalPointer ());
    showTable (item);
}

void MainWindow::showTable (QGraphicsItem * item)
{
    current_item = NULL; // disable on_tableWidget_itemChanged

    ui->tableWidget->clear ();
    ui->tableWidget->setColumnCount (2);
    ui->tableWidget->setHorizontalHeaderLabels (QStringList () << "Name" << "Value");
    ui->tableWidget->verticalHeader()->hide();

    ui->statusBar->showMessage ("point (1)");

    QObject * obj = dynamic_cast <QObject *> (item);
    if (obj != NULL)
    {
       const QMetaObject * cls = obj->metaObject ();
       assert (cls != NULL);

       int cnt = cls->propertyCount ();
       ui->tableWidget->setRowCount (cnt);
       for (int inx = 0; inx < cnt; inx ++)
       {
           QMetaProperty prop = cls->property (inx);
           QString attr = prop.name ();
           QVariant value = prop.read (obj);

           QTableWidgetItem * label = new QTableWidgetItem;
           label->setText (attr);
           ui->tableWidget->setItem (inx, 0, label);

           QTableWidgetItem * field = new QTableWidgetItem;
           field->setData (Qt::EditRole, value);
           ui->tableWidget->setItem (inx, 1, field);

           field->setData (Qt::UserRole, attr);
           field->setData (Qt::UserRole+1, inx);
           if (attr == "pen" or attr == "brush")
              field->setData (Qt::DecorationRole, value);
    }

    current_item = item;
    }
}

void MainWindow::on_tableWidget_itemChanged (QTableWidgetItem * item)
{
    if (current_item != NULL)
    {
       QString attr = item->data (Qt::UserRole).toString();
       int inx = item->data (Qt::UserRole+1).toInt();
       QVariant value = item->data (Qt::EditRole);
       ui->statusBar->showMessage ("update " + attr + " = " + value.toString ());

       QObject * obj = dynamic_cast <QObject *> (current_item);
       if (obj != NULL)
       {
          const QMetaObject * cls = obj->metaObject ();
          assert (cls != NULL);
          QMetaProperty prop = cls->property (inx);
          prop.write (obj, value);
       }

       // updateAttribute (current_item, attr, value);
    }
}

/* ---------------------------------------------------------------------- */

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    scene (NULL),
    current_item (NULL),
    top_item (NULL)
{
    ui->setupUi(this);

    global_win = this;

    ui->statusBar->showMessage(QString ("Qt ") + qVersion());

    /* setup toolbar */

    QString code = "<rectangle x='20' y='50' width='60' height='20' pen='red' brush='orange' />";
    ui->mainToolBar->addWidget (new ShapeButton ("rectangle", code));

    code = "<ellipse x='20' y='20' width='20' height='20' pen='red' brush='blue' />";
    ui->mainToolBar->addWidget (new ShapeButton ("ellipse", code));

    code = "<rectangle x='10' y='10' width='100' height='100' pen='red' brush='yellow'>"
           "  <rectangle x='20' y='50' width='60' height='20' pen='red' brush='orange'/>"
            "</rectangle>" ;
    ui->mainToolBar->addWidget (new ShapeButton ("block", code));

    QList <QString> names;
    names << "red" << "blue" << "green" << "yellow";

    for (auto name : names)
    {
       ColorButton * b = new ColorButton (name);
       ui->mainToolBar->addWidget (b);
    }

    /* setup scene */

    scene = new QGraphicsScene;
    ui->graphicsView->setScene (scene);

    scene->setSceneRect (0, 0, 1280, 1024);

    /* background */

    const int n = 16;
    QBitmap texture (n, n);
    texture.clear ();

    QPainter painter (& texture);
    painter.drawLine (0, 0, n-1, 0);
    painter.drawLine (0, 0, 0, n-1);

    QBrush brush (QColor ("cornflowerblue"));
    brush.setTexture (texture);
    scene->setBackgroundBrush (brush);

    /* plane */

    Rectangle * plane = new Rectangle;
    plane->setPen (QPen (Qt::NoPen));
    plane->setBrush (QBrush (Qt::NoBrush));
    plane->setFlags (0);
    plane->setRect (0, 0, 1280, 1024);

    top_item = plane;
    scene->addItem (top_item);

    /* initialization */

    readFile (":/data.xml");
    ui->treeView->setModel (new TreeModel (top_item));

    showTable (top_item->childItems()[0]);
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::on_actionQuit_triggered()
{
   this->close ()    ;
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
    QApplication a(argc, argv);
    MainWindow w;
    w.show();
    return a.exec();
}
